import math, random

import gym
import numpy as np

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F
from torch.distributions import Categorical

import sys
sys.path.append("..")
sys.path.append("../..")

import general_energy_nets.predictive_coding as pc

from IPython.display import clear_output
import matplotlib.pyplot as plt

import argparse


ENV_ID = "CartPole-v0"
NUM_SEEDS = 10

PC_LR = 0.01
PC_OPTIM = 'Adam'

BP_LR = 0.01
BP_OPTIM = 'Adam'

BATCH_SIZE = 16
PERCENTILE = 70

NESTEROV = False

MAX_ITERS = 100

class NN(nn.Module):
    def __init__(self, num_inputs, num_actions, predictive_coding=False):
        super(NN, self).__init__()
        
        self.layers = []
        self.layers.append(nn.Linear(num_inputs, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, 128))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(128, num_actions))
        self.layers = nn.Sequential(*self.layers)
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def act(self, state):
        self.train(False)
        state   = torch.FloatTensor(state).unsqueeze(0).to(device)
        logits = self.forward(state)
        probs = F.softmax(logits, dim=1)
        dist = Categorical(probs)
        action = dist.sample().cpu().numpy()[0]
        return action

def step_update(net, trainer, state, action):
    net.train()
    
    results = trainer.train_on_batch(
        inputs=state,
        loss_fn=cross_entropy_loss_fn,
        loss_fn_kwargs={
            '_targets': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results

def iter_batches(env, net, batch_size):
    batch = []
    episode_reward = 0.0
    episode_steps = []
    state = env.reset()
    
    net.eval()
    
    while True:
        action = net.act(state)
        
        next_state, reward, done, _ = env.step(action)
        episode_reward += reward
        
        episode_steps.append((state, action))
        
        if done:
            batch.append((episode_reward, episode_steps))
            episode_reward = 0.0
            episode_steps = []
            next_state = env.reset()
            if len(batch) == batch_size:
                yield batch
                batch = []
                
        state = next_state

def filter_batch(batch, percentile):
    rewards, steps = zip(*batch)
    reward_bound = np.percentile(rewards, percentile)
    reward_mean = float(np.mean(rewards))
    
    train_state = []
    train_act = []
    for reward, steps in batch:
        if reward < reward_bound:
            continue
            
        states, actions = zip(*steps)
        train_state.extend(states)
        train_act.extend(actions)

    return train_state, train_act, reward_bound, reward_mean

def cross_entropy_loss_fn(output, _targets):
    return nn.CrossEntropyLoss()(output, _targets)

def compute_confidence_intervals(matrix):
    upper_confidence_interval = []
    lower_confidence_interval = []

    for iteration in matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(iteration, len(iteration))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    return upper_confidence_interval, lower_confidence_interval

def plot_all(bp_reward_matrix, pc_reward_matrix, confidence_intervals=False):
    bp_reward_matrix = np.array(bp_reward_matrix)
    pc_reward_matrix = np.array(pc_reward_matrix)

    # plot BP
    plt.style.use('ggplot')
    bp_average_reward = np.mean(np.mean(bp_reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('{} : Cross Entropy with BP (Mean Reward = {:.2f})'.format(ENV_ID, bp_average_reward))
    bp_reward_matrix_mean = np.mean(bp_reward_matrix, axis=0)
    plt.plot(bp_reward_matrix_mean, linewidth=2, color='blueviolet', alpha=0.8)

    if confidence_intervals:
        bp_upper_confidence_interval, bp_lower_confidence_interval = compute_confidence_intervals(bp_reward_matrix)

    #plt.ylim([0, 500])
    plt.xlabel('iterations')
    plt.ylabel('reward')
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_upper_confidence_interval, bp_lower_confidence_interval, color='blueviolet', alpha=0.4)
    plt.savefig('./imgs/{}_bp_rewards_seeds={}_bp_lr={}_bp_optim={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, BP_LR, BP_OPTIM, NESTEROV))
    plt.close('all')

    # plot PC
    plt.style.use('ggplot')
    pc_average_reward = np.mean(np.mean(pc_reward_matrix, axis=0))
    plt.figure(figsize=(10, 5))
    plt.title('{} : Cross Entropy with PC (Mean Reward = {:.2f})'.format(ENV_ID, pc_average_reward))
    pc_reward_matrix_mean = np.mean(pc_reward_matrix, axis=0)
    plt.plot(pc_reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8)

    pc_upper_confidence_interval = []
    pc_lower_confidence_interval = []

    if confidence_intervals:
        pc_upper_confidence_interval, pc_lower_confidence_interval = compute_confidence_intervals(pc_reward_matrix)

    #plt.ylim([0, 500])
    plt.xlabel('iterations')
    plt.ylabel('reward')
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_upper_confidence_interval, pc_lower_confidence_interval, color='firebrick', alpha=0.4)
    plt.savefig('./imgs/{}_pc_rewards_seeds={}_pc_lr={}_pc_optim={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, PC_LR, PC_OPTIM, NESTEROV))
    plt.close('all')

    # plot both on same axis
    plt.style.use('ggplot')
    plt.figure(figsize=(10, 5))
    plt.title('{} : Cross Entropy method'.format(ENV_ID))
    plt.plot([x for x in range(bp_reward_matrix.shape[1])], bp_reward_matrix_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_reward_matrix.shape[1])], pc_reward_matrix_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()

    if confidence_intervals:
        plt.fill_between([x for x in range(bp_reward_matrix.shape[1])], bp_upper_confidence_interval, bp_lower_confidence_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_reward_matrix.shape[1])], pc_upper_confidence_interval, pc_lower_confidence_interval, color='firebrick', alpha=0.4)

    #plt.ylim([0, 500])
    plt.xlabel('iterations')
    plt.ylabel('reward')
    plt.savefig('./imgs/{}_seeds={}_batch_size={}_percentile={}_bp_lr={}_bp_optim={}_pc_lr={}_pc_lr={}_nesterov={}.png'.format(ENV_ID, NUM_SEEDS, BATCH_SIZE, PERCENTILE, BP_LR, BP_OPTIM, PC_LR, PC_OPTIM, NESTEROV))
    plt.close('all')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--env-id", default=ENV_ID)
    parser.add_argument("--batch-size", default=BATCH_SIZE)
    parser.add_argument("--percentile", default=PERCENTILE)
    parser.add_argument("--bp-lr", default=BP_LR)
    parser.add_argument("--pc-lr", default=PC_LR)
    parser.add_argument("--bp-optim", default=BP_OPTIM)
    parser.add_argument("--pc-optim", default=PC_OPTIM)
    parser.add_argument("--nesterov", default=NESTEROV)
    parser.add_argument("--max-iters", default=MAX_ITERS)
    args = parser.parse_args()

    NUM_SEEDS = int(args.n_seed)
    ENV_ID = args.env_id
    BATCH_SIZE = int(args.batch_size)
    PERCENTILE = int(args.percentile)
    BP_LR = float(args.bp_lr)
    PC_LR = float(args.pc_lr)
    BP_OPTIM = args.bp_optim
    PC_OPTIM = args.pc_optim
    NESTEROV = bool(int(args.nesterov))
    MAX_ITERS = int(args.max_iters)

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    # run BP
    bp_loss_matrix = []
    bp_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):
        env = gym.make(ENV_ID)
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        net = NN(env.observation_space.shape[0], env.action_space.n, predictive_coding=False).to(device)
        net.eval()

        optimizer = None
        p_kwargs = None
        if BP_OPTIM == 'Adam':
            optimizer = optim.Adam
            p_kwargs = {"lr": BP_LR}
        elif BP_OPTIM == 'SGD':
            optimizer = optim.SGD
            if NESTEROV:
                p_kwargs = {"lr": BP_LR, "momentum" : 0.9, "nesterov" : True}
            else:
                p_kwargs = {"lr": BP_LR}
        else:
            raise ValueError("Bad argument for optim")

        bp_trainer = pc.PCTrainer(
            net,
            T=1,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs=p_kwargs,
            plot_progress_at=[],
        )

        all_reward_means = []
        losses = []

        for eps_idx, batch in enumerate(iter_batches(env, net, BATCH_SIZE)):
            state, action, reward_bound, reward_mean = filter_batch(batch, PERCENTILE)
            
            state = torch.FloatTensor(np.float32(state)).to(device)
            action = torch.LongTensor(action).to(device)

            loss = step_update(net, bp_trainer, state, action)['loss'][0]
            losses.append(loss)
            
            all_reward_means.append(reward_mean)
            
            if eps_idx > MAX_ITERS:
                break

        print("Seed {} complete".format(seed))
        bp_loss_matrix.append(losses)
        bp_reward_matrix.append(all_reward_means)

    # run PC

    pc_loss_matrix = []
    pc_reward_matrix = []

    for seed in range(742, 742+NUM_SEEDS):
        env = gym.make(ENV_ID)
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        net = NN(env.observation_space.shape[0], env.action_space.n, predictive_coding=True).to(device)
        net.eval()

        optimizer = None
        p_kwargs = None
        if PC_OPTIM == 'Adam':
            optimizer = optim.Adam
            p_kwargs = {"lr": PC_LR}
        elif PC_OPTIM == 'SGD':
            optimizer = optim.SGD
            if NESTEROV:
                p_kwargs = {"lr": PC_LR, "momentum" : 0.9, "nesterov" : True}
            else:
                p_kwargs = {"lr": PC_LR}
        else:
            raise ValueError("Bad argument for optim")

        pc_trainer = pc.PCTrainer(
            net,
            T=128,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optimizer,
            optimizer_p_kwargs=p_kwargs,
            plot_progress_at=[],
        )

        all_reward_means = []
        losses = []

        for eps_idx, batch in enumerate(iter_batches(env, net, BATCH_SIZE)):
            state, action, reward_bound, reward_mean = filter_batch(batch, PERCENTILE)
            
            state = torch.FloatTensor(np.float32(state)).to(device)
            action = torch.LongTensor(action).to(device)

            loss = step_update(net, pc_trainer, state, action)['loss'][0]
            losses.append(loss)
            
            all_reward_means.append(reward_mean)
            
            if eps_idx > MAX_ITERS:
                break

        print("Seed {} complete".format(seed))
        pc_loss_matrix.append(losses)
        pc_reward_matrix.append(all_reward_means)

    plot_all(bp_reward_matrix, pc_reward_matrix, confidence_intervals=(NUM_SEEDS >= 3))