import math, random

import gym
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.font_manager

import sys
import torch
import torch.nn as nn
import torch.optim as optim
import torch.autograd as autograd 
import torch.nn.functional as F

import sys
sys.path.append("..")

import general_energy_nets.predictive_coding as pc

from value_iteration import value_iteration, plot_grid

from IPython.display import clear_output

from collections import deque

import argparse

# default number of seeds
NUM_SEEDS = 1
ENV_SIZE = 4
BP_LR = 0.1
PC_LR = 0.1
SYNC_FREQ = 10
BATCH_SIZE = 64
GAMMA = 0.7
CLAMP = True
HIDDEN_SIZE = 128

class ReplayBuffer(object):
    def __init__(self, capacity):
        self.buffer = deque(maxlen=capacity)
    
    def push(self, state, action, reward, next_state, done):
        state      = np.expand_dims(state, 0)
        next_state = np.expand_dims(next_state, 0)
            
        self.buffer.append((state, action, reward, next_state, done))
    
    def sample(self, batch_size):
        state, action, reward, next_state, done = zip(*random.sample(self.buffer, batch_size))
        return np.concatenate(state), action, reward, np.concatenate(next_state), done
    
    def __len__(self):
        return len(self.buffer)

def clamp_loss_fn(output, _targets):
    return (output - _targets).pow(2).sum() * 0.5

def loss_fn(output, _expected_q_value, _action):
    q_value = output.gather(1, _action.unsqueeze(1)).squeeze(1) 
    return (q_value - _expected_q_value).pow(2).sum() * 0.5

def update_target(current_model, target_model):
    parameters = current_model.state_dict()
    try:
        del parameters['layers.1._x']
        del parameters['layers.4._x']
    except:
        pass
    target_model.load_state_dict(parameters)

def compute_td_loss_with_clamp(batch_size, gamma):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = F.one_hot(torch.tensor(state).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
    next_state = F.one_hot(torch.tensor(next_state).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)
    
    # add random noise to the states
    
    #state = state + (torch.randn(state.shape) / 10.0).to(device)
    #new_state = new_state + (torch.randn(state.shape) / 10.0).to(device)
    
    #print(state.shape, next_state.shape)
    #print(action.shape, reward.shape, done.shape)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state) 
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()
        
        current_q_values = current_model(state)
        current_q_values = current_q_values.detach()
        
        batch_index = torch.arange(current_q_values.shape[0]).to(device)
        
    
    expected_q_value = reward + gamma * next_q_value
    current_q_values[batch_index, action] = expected_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=clamp_loss_fn,
        loss_fn_kwargs={
            '_targets': current_q_values
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results

def compute_td_loss_without_clamp(batch_size, gamma):
    state, action, reward, next_state, done = replay_buffer.sample(batch_size)

    state      = F.one_hot(torch.tensor(state).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
    next_state = F.one_hot(torch.tensor(next_state).to(torch.int64), num_classes=env.observation_space.n).to(torch.float32).to(device)
    action     = torch.LongTensor(action).to(device)
    reward     = torch.FloatTensor(reward).to(device)
    done       = torch.BoolTensor(done).to(device)

    #q_values      = current_model(state)
    #q_value       = q_values.gather(1, action.unsqueeze(1)).squeeze(1) 
    
    current_model.eval()
    
    with torch.no_grad():
        next_q_values = current_model(next_state)
        next_q_state_values = target_model(next_state)
        next_q_value = next_q_state_values.gather(1, torch.max(next_q_values, 1)[1].unsqueeze(1)).squeeze(1)
        next_q_value[done] = 0.0
        next_q_value = next_q_value.detach()

    expected_q_value = reward + gamma * next_q_value
    
    current_model.train()
    results = pc_trainer.train_on_batch(
        inputs=state,
        loss_fn=loss_fn,
        loss_fn_kwargs={
            '_expected_q_value': expected_q_value,
            '_action': action
        },
        is_log_progress=False,
        is_return_results_every_t=False,
        is_checking_after_callback_after_t=False,
    )
    
    return results

class DQN(nn.Module):
    def __init__(self, num_inputs, num_actions, hidden_size=128, predictive_coding=False):
        super(DQN, self).__init__()
        
        self.layers = []
        self.layers.append(nn.Linear(num_inputs, hidden_size))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(hidden_size, hidden_size))
        if predictive_coding:
            self.layers.append(pc.PCLayer())
        self.layers.append(nn.ReLU())
        self.layers.append(nn.Linear(hidden_size, num_actions))
        self.layers = nn.Sequential(*self.layers)
        
    def forward(self, x):
        return self.layers(x)
    
    @torch.no_grad()
    def get_policy_values(self):
        V = np.zeros(env.nS)
        policy = np.zeros([env.nS, env.nA]) / env.nA

        for s in range(env.nS):
            _s = F.one_hot(torch.tensor(s), num_classes=env.observation_space.n).to(torch.float32).unsqueeze(0).to(device)
            q_value = self.forward(_s)
            V[s] = q_value.max().cpu().item()
            policy[s] = F.one_hot(q_value.argmax().cpu(), num_classes=env.action_space.n).to(torch.float32)
            
        return policy, V
    
    @torch.no_grad()
    def act(self, state, epsilon):
        if random.random() > epsilon:
            state = F.one_hot(torch.tensor(state), num_classes=env.observation_space.n).to(torch.float32).unsqueeze(0).to(device)
            q_value = self.forward(state)
            action  = q_value.argmax().cpu().item()
        else:
            action = random.randrange(env.action_space.n)
        return action

def compute_policy_loss(optimal_policy, policy):
    # compute 0-1 policy loss
    policy_loss = 0.0
    
    for s in range(env.nS):
        # if the policy doesn't point in the same direction then it is not optimal
        if np.dot(policy[s], optimal_policy[s]) == 0.0:
            policy_loss += 1.0

    return policy_loss

def compute_value_loss(optimal_V, V):
    # compute mse value loss
    value_loss = 0.0
        
    return ((optimal_V - V)**2).sum()

def compute_confidence_intervals(matrix):
    upper_confidence_interval = []
    lower_confidence_interval = []

    for iteration in matrix.transpose():
        bootstrapped_means = []
        for B in range(2000):
            bootstrapped_sample = np.random.choice(iteration, len(iteration))
            bootstrapped_means.append(np.mean(bootstrapped_sample))
        bootstrapped_percentiles = np.percentile(np.array(bootstrapped_means), [2.5, 97.5])
        upper_confidence_interval.append(bootstrapped_percentiles[1])
        lower_confidence_interval.append(bootstrapped_percentiles[0])

    return upper_confidence_interval, lower_confidence_interval


def plot(bp_policy_loss_matrix, bp_value_loss_matrix, pc_policy_loss_matrix, pc_value_loss_matrix, confidence_intervals=False):
    plt.style.use('ggplot')

    max_iters = max([len(policy_losses) for policy_losses in bp_policy_loss_matrix])

    bp_policy_loss_matrix = np.array([(policy_losses + [policy_losses[-1]]*(max_iters - len(policy_losses))) for policy_losses in bp_policy_loss_matrix])
    bp_value_loss_matrix = np.array([(value_losses + [value_losses[-1]]*(max_iters - len(value_losses)))  for value_losses in bp_value_loss_matrix])

    bp_policy_loss_mean = np.mean(bp_policy_loss_matrix, axis=0)
    bp_value_loss_mean = np.mean(bp_value_loss_matrix, axis=0)

    max_iters = max([len(policy_losses) for policy_losses in pc_policy_loss_matrix])

    pc_policy_loss_matrix = np.array([policy_losses + [policy_losses[-1]]*(max_iters - len(policy_losses)) for policy_losses in pc_policy_loss_matrix])
    pc_value_loss_matrix = np.array([value_losses + [value_losses[-1]]*(max_iters - len(value_losses)) for value_losses in pc_value_loss_matrix])

    pc_policy_loss_mean = np.mean(pc_policy_loss_matrix, axis=0)
    pc_value_loss_mean = np.mean(pc_value_loss_matrix, axis=0)

    if confidence_intervals:
        bp_policy_loss_upper_interval, bp_policy_loss_lower_interval = compute_confidence_intervals(bp_policy_loss_matrix)
        bp_value_loss_upper_interval, bp_value_loss_lower_interval = compute_confidence_intervals(bp_value_loss_matrix)

        pc_policy_loss_upper_interval, pc_policy_loss_lower_interval = compute_confidence_intervals(pc_policy_loss_matrix)
        pc_value_loss_upper_interval, pc_value_loss_lower_interval = compute_confidence_intervals(pc_value_loss_matrix)

    # plot each individually
    # plot BP on its own
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('DQN-BP 0-1 policy losses')
    plt.plot([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_upper_interval, bp_policy_loss_lower_interval, color='blueviolet', alpha=0.4)

    plt.subplot(132)
    plt.title('DQN-BP MSE value losses')
    plt.plot([x for x in range(bp_value_loss_matrix.shape[1])], bp_value_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_value_loss_matrix.shape[1])], bp_value_loss_upper_interval, bp_value_loss_lower_interval, color='blueviolet', alpha=0.4)

    plt.savefig('./imgs/frozen_lake{}x{}_bp_seeds={}_sync_freq={}_batch_size={}_gamma={:.2f}_bp_lr={}.png'.format(ENV_SIZE, ENV_SIZE, NUM_SEEDS, SYNC_FREQ, BATCH_SIZE, GAMMA, BP_LR))
    plt.close('all')

    # plot PC on its own
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('DQN-PC-C 0-1 policy losses')
    plt.plot([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_upper_interval, pc_policy_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.subplot(132)
    plt.title('DQN-PC-C MSE value losses')
    plt.plot([x for x in range(pc_value_loss_matrix.shape[1])], pc_value_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    if confidence_intervals:
        plt.fill_between([x for x in range(pc_value_loss_matrix.shape[1])], pc_value_loss_upper_interval, pc_value_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.savefig('./imgs/frozen_lake{}x{}_pc_seeds={}_sync_freq={}_batch_size={}_gamma={:.2f}_clamp={}_pc_lr={}.png'.format(ENV_SIZE, ENV_SIZE, NUM_SEEDS, SYNC_FREQ, BATCH_SIZE, GAMMA, CLAMP, PC_LR))
    plt.close('all')

    # plot both on same axis
    plt.figure(figsize=(20,5))
    plt.subplot(131)
    plt.title('0-1 policy losses')
    plt.plot([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()
    
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_policy_loss_matrix.shape[1])], bp_policy_loss_upper_interval, bp_policy_loss_lower_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_policy_loss_matrix.shape[1])], pc_policy_loss_upper_interval, pc_policy_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.subplot(132)
    plt.title('MSE value losses')
    plt.plot([x for x in range(bp_value_loss_matrix.shape[1])], bp_value_loss_mean, linewidth=2, color='blueviolet', alpha=0.8, label="BP")
    plt.plot([x for x in range(pc_value_loss_matrix.shape[1])], pc_value_loss_mean, linewidth=2, color='firebrick', alpha=0.8, label="PC")
    plt.legend()
    
    if confidence_intervals:
        plt.fill_between([x for x in range(bp_value_loss_matrix.shape[1])], bp_value_loss_upper_interval, bp_value_loss_lower_interval, color='blueviolet', alpha=0.4)
        plt.fill_between([x for x in range(pc_value_loss_matrix.shape[1])], pc_value_loss_upper_interval, pc_value_loss_lower_interval, color='firebrick', alpha=0.4)

    plt.savefig('./imgs/frozen_lake{}x{}_seeds={}_sync_freq={}_batch_size={}_gamma={:.2f}_clamp={}_bp_lr={}_pc_lr={}.png'.format(ENV_SIZE, ENV_SIZE, NUM_SEEDS, SYNC_FREQ, BATCH_SIZE, GAMMA, CLAMP, BP_LR, PC_LR))
    plt.close('all')

def save_data(bp_policy_loss_matrix, pc_policy_loss_matrix, bp_value_loss_matrix, pc_value_loss_matrix):
    bp_mean_time_to_solve = np.mean([len(policy_losses) for policy_losses in bp_policy_loss_matrix])
    pc_mean_time_to_solve = np.mean([len(policy_losses) for policy_losses in pc_policy_loss_matrix])

    print('Experiment done : BP average solve time {:.2f}, PC average solve time {:.2f}'.format(bp_mean_time_to_solve, pc_mean_time_to_solve))

    max_iterations = 1000 if ENV_SIZE == 4 else 4000

    bp_mean_policy_loss = np.mean(np.sum(bp_policy_loss_matrix, axis=0)) * bp_mean_time_to_solve / max_iterations
    pc_mean_policy_loss = np.mean(np.mean(pc_policy_loss_matrix, axis=0)) * pc_mean_time_to_solve / max_iterations

    bp_mean_value_loss = np.mean(np.mean(bp_value_loss_matrix, axis=0)) * bp_mean_time_to_solve / max_iterations
    pc_mean_value_loss = np.mean(np.mean(pc_value_loss_matrix, axis=0)) * pc_mean_time_to_solve / max_iterations

    with open('./data/frozen_lake{}x{}_seeds={}_sync_freq={}_batch_size={}_gamma={:.2f}_clamp={}_bp_lr={}_pc_lr={}.txt'.format(ENV_SIZE, ENV_SIZE, NUM_SEEDS, SYNC_FREQ, BATCH_SIZE, GAMMA, CLAMP, BP_LR, PC_LR), 'w') as outfile:
        outfile.write('BP average solve time {:.2f}\n'.format(bp_mean_time_to_solve))
        outfile.write('PC average solve time {:.2f}\n'.format(pc_mean_time_to_solve))
        outfile.write('BP Mean Policy Loss ({}) : \n'.format(np.mean(bp_mean_policy_loss)))
        outfile.write(str(bp_mean_policy_loss))
        outfile.write('\n')
        outfile.write('PC Mean Policy Loss ({}) : \n'.format(np.mean(pc_mean_policy_loss)))
        outfile.write(str(pc_mean_policy_loss))
        outfile.write('\n')
        outfile.write('BP Mean Value Loss ({}) : \n'.format(np.mean(bp_mean_value_loss)))
        outfile.write(str(bp_mean_value_loss))
        outfile.write('\n')
        outfile.write('PC Mean Value Loss ({}) : \n'.format(np.mean(pc_mean_value_loss)))
        outfile.write(str(pc_mean_value_loss))
        outfile.write('\n')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--n-seed", default=NUM_SEEDS)
    parser.add_argument("--env-size", default=ENV_SIZE)
    parser.add_argument("--bp-lr", default=BP_LR)
    parser.add_argument("--pc-lr", default=PC_LR)
    parser.add_argument("--sync-freq", default=SYNC_FREQ)
    parser.add_argument("--batch-size", default=BATCH_SIZE)
    parser.add_argument("--gamma", default=GAMMA)
    parser.add_argument("--clamp", default=CLAMP)
    parser.add_argument("--hidden-size", default=HIDDEN_SIZE)
    
    args = parser.parse_args()
    
    NUM_SEEDS = int(args.n_seed)
    ENV_SIZE = int(args.env_size)
    BP_LR = float(args.bp_lr)
    PC_LR = float(args.pc_lr)
    SYNC_FREQ = int(args.sync_freq)
    BATCH_SIZE = int(args.batch_size)
    GAMMA = float(args.gamma)
    CLAMP = bool(int(args.clamp))
    HIDDEN_SIZE = int(args.hidden_size)

    assert ENV_SIZE in [4, 8]

    if ENV_SIZE == 4:
        assert BATCH_SIZE <=64
    else:
        assert BATCH_SIZE <= 256

    assert torch.cuda.is_available()
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu" )

    env_name = 'FrozenLake-v1' if ENV_SIZE == 4 else 'FrozenLake8x8-v1'
    env = gym.make(env_name, is_slippery=False) 

    # compute the optimal policy and values
    policy_optim, V_optim = value_iteration(env, gamma=GAMMA)

    # set up the replay buffer
    replay_buffer = ReplayBuffer(64 if ENV_SIZE == 4 else 256)

    # add every possible (state, action, reward, next_state, done) tuple to the replay buffer by iterating over state-action pairs
    for state in range(env.nS):
        for action in range(env.nA):
            for _, next_state, reward, done in env.P[state][action]:
                replay_buffer.push(state, action, reward, next_state, done)

    # run BP
    bp_policy_loss_matrix = []
    bp_value_loss_matrix = []

    for seed in range(742, 742+NUM_SEEDS):

        env = gym.make(env_name, is_slippery=False) 

        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        current_model = DQN(env.observation_space.n, env.action_space.n, hidden_size=HIDDEN_SIZE).to(device)
        target_model  = DQN(env.observation_space.n, env.action_space.n, hidden_size=HIDDEN_SIZE).to(device)
        target_model.eval()
        
        target_model.load_state_dict(current_model.state_dict())
          
        pc_trainer = pc.PCTrainer(
            current_model,
            T=1,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optim.SGD,
            optimizer_p_kwargs={"lr": BP_LR},
            plot_progress_at=[],
        )

        num_iterations = 1000 if ENV_SIZE == 4 else 4000
        batch_size = BATCH_SIZE
        gamma      = GAMMA
        sync_freq = SYNC_FREQ

        policy_losses = []
        value_losses = []
        losses = []

        for iteration in range(1, num_iterations + 1):
            loss = compute_td_loss_without_clamp(batch_size, gamma)['loss'][0]
            losses.append(loss)
                
            current_model.eval()
            policy, V = current_model.get_policy_values()
            policy_loss = compute_policy_loss(policy_optim, policy)
            value_loss = compute_value_loss(V_optim, V)
            policy_losses.append(policy_loss)
            value_losses.append(value_loss)
                
            if iteration % sync_freq == 0:
                target_model.load_state_dict(current_model.state_dict())
            
            if policy_losses[-1] == 0.0 and value_losses[-1] < 1e-4 or (iteration == num_iterations):
                # solved the environment
                print("Iteration {} batch loss {} policy loss {} value loss {}".format(
                    iteration, np.mean(losses[-10:]), np.mean(policy_losses[-10:]), np.mean(value_losses[-10:])
                ))
                
                break

        bp_policy_loss_matrix.append(policy_losses)
        bp_value_loss_matrix.append(value_losses)

    # run PC

    pc_policy_loss_matrix = []
    pc_value_loss_matrix = []

    for seed in range(742, 742+NUM_SEEDS):

        env = gym.make(env_name, is_slippery=False) 

        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        env.action_space.seed(seed)
        env.seed(seed)

        current_model = DQN(env.observation_space.n, env.action_space.n, hidden_size=HIDDEN_SIZE, predictive_coding=True).to(device)
        target_model  = DQN(env.observation_space.n, env.action_space.n, hidden_size=HIDDEN_SIZE, predictive_coding=True).to(device)
        target_model.eval()
        
        update_target(current_model, target_model)

        pc_trainer = pc.PCTrainer(
            current_model,
            T=128,
            update_x_at='all',
            optimizer_x_fn=optim.SGD,
            optimizer_x_kwargs={"lr": 0.1},
            x_lr_discount=0.5,
            update_p_at='last',
            optimizer_p_fn=optim.SGD,
            optimizer_p_kwargs={"lr": PC_LR},
            plot_progress_at=[],
        )

        num_iterations = 1000 if ENV_SIZE == 4 else 4000
        batch_size = BATCH_SIZE
        gamma      = GAMMA
        sync_freq = SYNC_FREQ

        policy_losses = []
        value_losses = []
        losses = []

        for iteration in range(1, num_iterations + 1):
            if CLAMP:
                loss = compute_td_loss_with_clamp(batch_size, gamma)['loss'][0]
            else:
                loss = compute_td_loss_without_clamp(batch_size, gamma)['loss'][0]
            losses.append(loss)
                
            current_model.eval()
            policy, V = current_model.get_policy_values()
            policy_loss = compute_policy_loss(policy_optim, policy)
            value_loss = compute_value_loss(V_optim, V)
            policy_losses.append(policy_loss)
            value_losses.append(value_loss)
                
            if iteration % sync_freq == 0:
                update_target(current_model, target_model)
            
            if policy_losses[-1] == 0.0 and value_losses[-1] < 1e-4 or (iteration == num_iterations):
                # solved the environment
                print("Iteration {} batch loss {} policy loss {} value loss {}".format(
                    iteration, np.mean(losses[-10:]), np.mean(policy_losses[-10:]), np.mean(value_losses[-10:])
                ))
                
                break

        pc_policy_loss_matrix.append(policy_losses)
        pc_value_loss_matrix.append(value_losses)

    plot(bp_policy_loss_matrix, bp_value_loss_matrix, pc_policy_loss_matrix, pc_value_loss_matrix, confidence_intervals=(NUM_SEEDS >= 10))
    save_data(bp_policy_loss_matrix, pc_policy_loss_matrix, bp_value_loss_matrix, pc_value_loss_matrix)
        